#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <vector>

using namespace std;

static const unsigned XDIM = 512;  //800;       //700
static const unsigned YDIM = 480; //XDIM*9/16; //512

#define INITIAL_W 0
#define USEMANDMAP 1    // cool clippings
#define DRAWDOUBLEPIX 0
#define DISPLAYRENDER 0
#define PROFILEMINMAX 0
#define SPHERETEST 0
#define MAXITER 8
#define BEAMSPEED 0.01
#define PRECISION 0.00001
#define DISTORTION DEG2RAD(0.5)
#define PROJSIZEX (XDIM/(double)YDIM)
#define PROJSIZEY 1.0
#define PROJDIST 0.7
#define EYEDIST  5.5
#define MAXTRACE     ((EYEDIST+2)/BEAMSPEED)
#define MAXTRACESTR "((EYEDIST+2)/BEAMSPEED)"
#define MIRRORCOUNT 4
#define AMBIENT     (0.0, 0.0, 0.0)
#define AMBIENTSTR "(0.0, 0.0, 0.0)"
#define TRACEPOW 19
#define SKYPOW 14

#define DEG2RAD(r) ((r) * M_PI / 180.0)

#define FLOAT double
#define FLOAT_ARG FLOAT
//#define FLOAT_ARG const FLOAT &
#define SIN sin
#define COS cos
#define SQRT sqrt
#define POW pow
#define PF "f"

#include "xyz.inc"
#include "matrix.inc"

const XYZ XYZ::RotatedAround(const XYZ &axis, FLOAT_ARG angle) const
{
	Matrix m(axis, angle);
	return RotatedBy(m);
}

class RGB
{
protected:
    FLOAT r,g,b;
private:    
    inline const int intensity(FLOAT elem) const
    {
        if(elem < 0)return 0;
        const FLOAT maxi = max(max(r,g), b);
        if(maxi <= 1)return int(255*elem);
        if(elem > 1)return 255;
        const FLOAT jaettavaa = maxi-1;
        int v = int((elem + jaettavaa)*255);
        return v>=255 ? 255 : v;
    }
    
public:
    RGB() : r(0), g(0), b(0)
    {
    }
    RGB(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B) : r(R), g(G), b(B)
    {
    }
    const RGB &operator *=(const RGB &p) { r*=p.r; g*=p.g; b*=p.b; return *this; }
    const RGB &operator +=(const RGB &p) { r+=p.r; g+=p.g; b+=p.b; return *this; }
    const RGB &operator -=(const RGB &p) { r-=p.r; g-=p.g; b-=p.b; return *this; }
    const RGB &operator *=(FLOAT_ARG p) { r*=p; g*=p; b*=b; return *this; }
    const RGB &operator /=(FLOAT_ARG p) { r/=p; g/=p; b/=b; return *this; }
    const RGB operator *(const RGB &p) { return RGB(r*p.r, g*p.g, b*p.b); }
    const RGB operator +(const RGB &p) { return RGB(r+p.r, g+p.g, b+p.b); }
    const RGB operator -(const RGB &p) { return RGB(r-p.r, g-p.g, b-p.b); }
    const RGB operator +(FLOAT_ARG p) { return RGB(r+p, g+p, b+p); }
    const RGB operator -(FLOAT_ARG p) { return RGB(r-p, g-p, b-p); }
    const RGB operator *(FLOAT_ARG p) { return RGB(r*p, g*p, b*p); }
    const RGB operator /(FLOAT_ARG p) { return RGB(r/p, g/p, b/p); }
    
    const unsigned char getR() const { return intensity(r); }
    const unsigned char getG() const { return intensity(g); }
    const unsigned char getB() const { return intensity(b); }
    
    const FLOAT power() const { return r+g+b; }
    
    void debug()
    {
//      printf("Returning %.2f %.2f %.2f\n", r,g,b);
    }
};

static class Pigment : public RGB
{
public:
    FLOAT diffmul, dirdiffmul, specmul, penetration;

	Pigment()
	{
		setdefaultmultipliers();
	}
	Pigment(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B) : RGB(R,G,B)
	{
		setdefaultmultipliers();
	}
	void set(FLOAT_ARG R, FLOAT_ARG G, FLOAT_ARG B)
	{
		r=R;
		g=G;
		b=B;
	}
	
    void setdefaultmultipliers()
    {
    	diffmul   = 0.7;
        dirdiffmul= 0.0;
        specmul   = 0.9;
        penetration=0.0;
    }
    void setmultipliers(FLOAT_ARG diff, FLOAT_ARG dirdiff, FLOAT_ARG spec)
    {
    	diffmul = diff;
    	dirdiffmul = dirdiff;
    	specmul = spec;
    }
} pigment;

static class Fract
{
public:
     /* Nm vakiot kuuluvat Fract-luokalle, mutta kun ansi c++ ei salli
     * mritt static const doubleja luokalle. static const inttej vain.
     * GNU C++ selviisi kyll niist, mutta en jaksa katsella niit varoituksia...
     * Update: ovat nyt sittenkin tll. Koodista tulisi sotkuista ilman nit
     * luokan sisll. Poistan varoituksen kntjn optiolla.
     */
    static const FLOAT xmin = -2.0;
    static const FLOAT xmax = +2.0;
    static const FLOAT ymin = -1.3;
    static const FLOAT ymax = +1.3;
    static const FLOAT rmin = -2.7;
    static const FLOAT rmax = +0.5;
    static const FLOAT imin = -1.13;
    static const FLOAT imax = +1.13;
#if PROFILEMINMAX
    FLOAT xmi,xma, ymi,yma, rmi,rma, imi,ima;
#endif
public:
    Fract()
#if PROFILEMINMAX
    : xmi(xmax),xma(xmin),
              ymi(ymax),yma(ymin),
              rmi(rmax),rma(rmin),
              imi(imax),ima(imin)
#endif
    {
#if USEMANDMAP
        FLOAT ip = imin;
        const FLOAT iadd=(imax-imin) / YDIM;
        const FLOAT radd=(rmax-rmin) / XDIM;
        for(unsigned y=0; y<YDIM; ++y)
        {
            FLOAT rp = rmin;
            for(unsigned x=0; x<XDIM; ++x)
            {
                mandelmap[y][x] = ismandel(rp, ip);
                rp += radd;
            }
            ip += iadd;
        }
#endif
    }
    
private:
#if USEMANDMAP
    bool mandelmap[YDIM][XDIM];
#endif
    inline bool ismandel(FLOAT_ARG RP, FLOAT_ARG IP)
    {
        if(IP<imin || IP>imax
        || RP<rmin || RP>rmax)return false;
        register FLOAT r = RP, i = IP;
        for(register int c=0; ; )
        {
            register const FLOAT r2 = r*r;
            register const FLOAT i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }

    inline bool isjulia(FLOAT_ARG XP, FLOAT_ARG YP, FLOAT_ARG RP, FLOAT_ARG IP)
    {
    	if(IP < -2.4)
    	{
    		/* mirror plate */
	    	pigment.set(1.00, 1.00, 1.00);
	    	pigment.setmultipliers(0.0, 0.06, 1.0);
        	return true;
    	}
    	
#if SPHERETEST
        pigment = RGB(0.9, 0.2, 0.7);
        
        if((XP*XP + (RP-1.3)*(RP-1.3) + IP*IP) < 1.4)return true;
        if((XP*XP + (RP+1.3)*(RP+1.3) + IP*IP) < 1.4)return true;
        if(((XP-2.3)*(XP-2.3) + (RP)*(RP) + (IP-2.5)*(IP-2.5)) < 0.8)return true;
        if(((XP+1.6)*(XP+1.6) + (RP)*(RP) + (IP+1.2)*(IP+1.2)) < 0.6)return true;
        return false;
#else        
        if(((XP-2.3)*(XP-2.3) + (RP)*(RP) + (IP-2.5)*(IP-2.5)) < 0.8
        || ((XP+1.6)*(XP+1.6) + (RP)*(RP) + (IP+1.2)*(IP+1.2)) < 0.6)
        {
	    	pigment.set(0.2, 0.6, 0.95);
	    	pigment.setmultipliers(0.7, 0.0, 0.8);
        	return true;
        }

        if(wayout(XP,YP, RP,IP))return false;
#if USEMANDMAP
        if(!mandelmap[unsigned((IP-imin)*YDIM/(imax-imin))]
                     [unsigned((RP-rmin)*XDIM/(rmax-rmin))])return false;
#endif
        //if(((XP)*(XP) + (RP+1)*(RP+1) + (IP-0.6)*(IP-0.6)) < 1)return true;
        
        register FLOAT r = XP, i = YP;
        for(register int c=0; ; )
        {
            register const FLOAT r2 = r*r;
            register const FLOAT i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
#if PROFILEMINMAX
                if(XP<xmi){xmi=XP;printf("New X min: %.6f\n", xmi);}
                if(YP<ymi){ymi=YP;printf("New Y min: %.6f\n", ymi);}
                if(RP<rmi){rmi=RP;printf("New R min: %.6f\n", rmi);}
                if(IP<imi){imi=IP;printf("New I min: %.6f\n", imi);}
                if(XP>xma){xma=XP;printf("New X max: %.6f\n", xma);}
                if(YP>yma){yma=YP;printf("New Y max: %.6f\n", yma);}
                if(RP>rma){rma=RP;printf("New R max: %.6f\n", rma);}
                if(IP>ima){ima=IP;printf("New I max: %.6f\n", ima);}
#endif
                pigment.set(0.9, 0.2, 0.7);
		    	pigment.setmultipliers(0.7, 0.0, 0.9);
        	
                return true;
            }
        }
#endif
    }
    
public:
    /* Paluuarvot: true=sisll, false=ulkona */
    inline bool hit(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        return isjulia(x,y,r,i);
    }
    
    /* Paluuarvot: true=tosi ulkona, false=ehk sisll */
    bool wayout(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        /* Mandelbrot-rajat */
        if(i<imin || i>imax
        || r<rmin || r>rmax)return true;

        /* Julia-rajat */
        if(x<xmin || x>xmax
        || y<ymin || y>ymax)return true;
        
        return false;
    }
} Fract;

class Light
{
public:
    XYZ pos, oldpos;
    RGB type;
    FLOAT rad;
    
    Light(const XYZ &p, const RGB &c, FLOAT_ARG r) : pos(p), type(c), rad(r)
    {
        oldpos = pos;
    }
};

static vector<Light> Lights;

static const FLOAT w = INITIAL_W; /* the time axis position */

/* return value: false=mahiksia, true=ulos menossa */
static bool NoChances(const XYZ &beam, const XYZ &dir)
{
    if((beam.z < Fract.xmin && (beam.z+dir.z*MAXTRACE < Fract.xmin))
    || (beam.x < Fract.rmin && (beam.x+dir.x*MAXTRACE < Fract.rmin))
    || (beam.y < Fract.imin && (beam.y+dir.y*MAXTRACE < Fract.imin))
    || (w      < Fract.ymin              )
    
    || (beam.z > Fract.xmax && (beam.z+dir.z*MAXTRACE > Fract.xmax))
    || (beam.x > Fract.rmax && (beam.x+dir.x*MAXTRACE > Fract.rmax))
    || (beam.y > Fract.imax && (beam.y+dir.y*MAXTRACE > Fract.imax))
    || (w      > Fract.ymax              )
    )
        return true;

    return false;
}

static void SkipEmpty(XYZ &beam, const XYZ &dir, unsigned &step)
{
#if 0
    printf("z %.6f += %.6f\n"
           "x %.6f += %.6f\n"
           "y %.6f += %.6f\n",
           beam.z, dir.z,
           beam.x, dir.x,
           beam.y, dir.y);
#endif           
    #if 1
    
    FLOAT zspace = 0;
    FLOAT wspace = 0;
    FLOAT xspace = 0;
    FLOAT yspace = 0;

    if(beam.z < Fract.xmin && dir.z > 0)zspace = Fract.xmin - beam.z; else if(beam.z > Fract.xmax && dir.z < 0)zspace = beam.z - Fract.xmax;
    if(beam.x < Fract.rmin && dir.x > 0)xspace = Fract.rmin - beam.x; else if(beam.x > Fract.rmax && dir.x < 0)xspace = beam.x - Fract.rmax;
    if(beam.y < Fract.imin && dir.y > 0)yspace = Fract.imin - beam.y; else if(beam.y > Fract.imax && dir.y < 0)yspace = beam.y - Fract.imax;
    if(w      < Fract.ymin             )wspace = Fract.ymin - w     ; else if(w      > Fract.ymax             )wspace = w      - Fract.ymax;
    
    FLOAT minimi = SQRT(zspace*zspace + wspace*wspace + xspace*xspace + yspace*yspace) - BEAMSPEED;
    //FLOAT minimi = max(max(zspace,wspace), max(xspace,yspace)) - BEAMSPEED;
    
    #if 0
    for(unsigned l=0; l<Lights.size(); ++l)
    {
        FLOAT dist = beam.dist(Lights[l].pos);
        if(dist < minimi)minimi = dist;
    }
    #endif
    if(minimi >= BEAMSPEED)
    {
        FLOAT add = minimi / dir.length();
        step += int(add);
        beam += dir * add;
    }
    #endif
}

// Tarkoitus on muuttaa beamia, mutta ei dirri.
static void FocusHit(XYZ &beam, XYZ dir)
{
    /// Find the most exact position of hit.
    FLOAT prec = BEAMSPEED;
    if(prec > PRECISION)
    {
        /* Beam points to stuff. */
        beam -= dir;
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
        	/* We are INSIDE the stuff!
        	 * Do not process further, this might be an infinite surface.
        	 */
        	return;
        }
        for(;;)
        {
            /* Beam points to space. */
            dir /= 2;
            prec /= 2;
            beam += dir;
            
            /* Check where does the beam point now. */            
            if(Fract.hit(beam.z, w, beam.x, beam.y))
            {
                /* Enough precision? */
                if(prec <= PRECISION)return;
                /* Back, want more precision */
                beam -= dir;
            }
            /* Continue */
        }
    }
    /* Done */
}

static void RelativeTrace(XYZ &beam, const XYZ &dir)
{
    /* Jos ollaan sisll, */
    if(Fract.hit(beam.z, w, beam.x, beam.y))
    {
        /* Kaivaudutaan ulos */
        for(unsigned max=0; ; )
        {
            beam -= dir;
            if(!Fract.hit(beam.z, w, beam.x, beam.y))break;
            
            if(++max >= MAXTRACE)return;
        }
        /* Takaisin sislle */
        beam += dir;
    }
    else
    {
        /* Jos taas ulkona, trmtn sislle */
        for(unsigned max=0; ; )
        {
            beam += dir;
            if(Fract.hit(beam.z, w, beam.x, beam.y))break;
            
            if(++max >= MAXTRACE)
                return;
        }
    }
    /* Sen jlkeen etsitn tarkka osumakohta */
    FocusHit(beam, dir);
}

/*
    tracetypes:
    -1  = find light
     0  = find surface, tracetype = bouncecount
*/
static RGB RayTrace(const XYZ &startpoint, const XYZ &dir, int tracetype)
{
    XYZ beam = startpoint;
    
    unsigned step=0;    
    
    for(; ; ++step)
    {
        if(tracetype == -1)
        {
            /// Trying to find light?
            /// Then check if we are within the light source radius.
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                FLOAT dist = beam.dist(Lights[l].pos);
                if(dist < Lights[l].rad)
                {
                    /// Inside the light source.
                	goto NoChances;
                }
            }
        }
        
#if 0
    #if !SPHERETEST
        if(tracetype >= 0 && NoChances(beam, dir))
        {
        	goto NoChances;
        }
        if(!(step % (unsigned)(MAXTRACE/30)))
            SkipEmpty(beam, dir, step);
    #endif
#endif
    
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
            if(tracetype < 0)
            {
                /// We were trying to find light.
                /// Instead, we hit a surface.
                /// So we found no light.
                return RGB AMBIENT;
            }
            
            FocusHit(beam, dir);
            
            XYZ randvec(0,0,1);
            if(dir.x==0.0)randvec.x=1;
            XYZ xdir = randvec.CrossProduct(dir);
            XYZ dir1 = dir.RotatedAround(xdir, DISTORTION);
            XYZ dir2 = dir1.RotatedAround(dir, DEG2RAD(240));
            XYZ dir3 = dir1.RotatedAround(dir, DEG2RAD(120));
            
            FLOAT surfacedistance = beam.dist(startpoint);
            
            XYZ end1 = startpoint + dir1 * (surfacedistance); RelativeTrace(end1, dir1);
            XYZ end2 = startpoint + dir2 * (surfacedistance); RelativeTrace(end2, dir2);
            XYZ end3 = startpoint + dir3 * (surfacedistance); RelativeTrace(end3, dir3);
            
            XYZ normal = (end2-end1).CrossProduct(end3-end1);
            normal.Normalize();

            /// We have now hit a solid surface.
            
            /// Declare the colour as it appears in the current lighting.            
            RGB result AMBIENT;
            Pigment thispigment = pigment;
            
            /// Now check all the light sources
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                XYZ live = Lights[l].pos - beam;
                live.Normalize();
                
                FLOAT effect = normal.DotProduct(live);
                if(effect <= 0)continue;
                if(thispigment.diffmul != 0.0)
                {
                    /* Seek for something blocking the light, or then add it. */
                    XYZ tmp = live * BEAMSPEED;
                    result += RayTrace(beam+tmp, tmp, -1) * (thispigment.diffmul * effect);
                }
                if(thispigment.dirdiffmul != 0.0)
                {
                    /* Draw the light, whether it's seen or not */
                    /* This ignores all possible shadows. */
                    result += Lights[l].type * (effect * thispigment.dirdiffmul);
                }
            }
            
            if(thispigment.specmul != 0.0)
            {
                XYZ reversevec = -dir;
                XYZ mirrored = reversevec.MirroredAround(normal);
                
                /*dir.print();
                printf("\nmirrored around "); normal.print();
                printf("\nis: "); mirrored.print();
                printf("\n\n");*/
                
                /* Find light in this mirror direction */
                result += RayTrace(beam+mirrored, mirrored, -1) * (thispigment.specmul);
                
                /* Find mirror surfaces too, if wanted */
                if(tracetype > 0)
                {
                    result += RayTrace(beam+mirrored, mirrored, tracetype - 1) * thispigment.specmul;
                }
            }
            
            /// The pigment affects all light sources.            
            result *= thispigment;
            
            result.debug();
            return result;
        }

        if(step >= MAXTRACE)
        {
        	// No chances
        	break;
        }
        
        beam += dir;
    }
    
NoChances:
    /* Found *nothing* in this direction */
    RGB result(0,0,0);
    XYZ yksdir(dir);
    yksdir.Normalize();
    
    /* Calculate some halo */
    for(unsigned l=0; l<Lights.size(); ++l)
    {
        XYZ live = Lights[l].pos - startpoint;
        live.Normalize();
        
        FLOAT effect = yksdir.DotProduct(live);
        if(effect <= 0)continue;
        unsigned potens = (tracetype<0) ? TRACEPOW : SKYPOW;
        result += Lights[l].type * POW(effect, potens);
    }
    
    return result;
}

static int ALKUY=0;
#if DISPLAYRENDER
#include <linux/fb.h>
static class Video
{
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        /* This palette is RGB and used by XAWTV ;) */
        const int r=5;
        const int g=9;
        const int b=5;
        const int rs = 256 / (r - 1);
        const int gs = 256 / (g - 1);
        const int bs = 256 / (b - 1);
        for (int i = 0; i < r*g*b; i++)
        {
            cmap.green[i+16] = (gs * ((i / (r * b)) % g)) * 255;
            cmap.red[i+16]   = (rs * ((i / b) % r)) * 255;
            cmap.blue[i+16]  = (bs * ((i) % b)) * 255;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        y += ALKUY;
        #if DRAWDOUBLEPIX
        if(XDIM<1280/2)x+=(1280-XDIM*2)/4;
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
        #else
        p[y*xres+x] = c;
        #endif
    }
} PSet;
#endif


int main(int argc, const char *const *argv)
{
printf("  Heijastuslhteiden laskennan rekursioraja: %d\n"
       "  Ambient-valo: %s\n"
       "\n"
       "Fraktaalin parametrit:\n"
       "  Iteraatiomr: %d\n"
       "  Akselit: mandelbrot{z,w}, julia{x,y} jossa w=%.2"PF"\n"
       "\n"
       "Ja muut laskentaparametrit:\n"
       "  Kuvasuhde: %u/%u\n"
       "  Harppausvli: %.4f\n"
       "  Pinnan tarkkuus: %.8f\n"
       "  Kameran etisyys origosta: %.2f\n"
       "  Projektiopinnan etisyys kamerasta: %.2f\n"
       "  Maksimiharppausmr: %s (= %u)\n"
       "\n",
       MIRRORCOUNT, 
       AMBIENTSTR,
       MAXITER,
       w,
       XDIM, YDIM,
       BEAMSPEED,
       PRECISION,
       EYEDIST,
       PROJDIST,
       MAXTRACESTR,
       (unsigned)(MAXTRACE));

//	Lights.push_back(Light(XYZ(+0.0, -3.6, +1.0),   RGB(0.30, 0.10, 1.10), 0.3));   // Violetti taustavalo
//	Lights.push_back(Light(XYZ(-3.0, +3.3, -2.4),   RGB(1.00, 0.40, 0.10), 0.2));   // Punainen nurkkavalo
//	Lights.push_back(Light(XYZ(-1.0, -1.0, -2.0),   RGB(0.30, 0.70, 0.20), 0.1)); // Vaaleanvihre valo...
//	Lights.push_back(Light(XYZ(+0.0, +0.0, -2.6),   RGB(1.00, 1.00, 1.00), 1.0)); // Valkoinen spottiheijastusvalo
//	Lights.push_back(Light(XYZ(+2.0, +4.0, +1.0),   RGB(0.10, 0.60, 1.00), 0.4));   // Sininen nurkkavalo
//	Lights.push_back(Light(XYZ(+4.0, +0.0, -0.0),   RGB(0.30, 0.10, 1.10), 2.3));   // Sininen sivuvalo
//	Lights.push_back(Light(XYZ(+0.0, +0.0, -7.0),   RGB(0.30, 0.10, 1.10), 2.3));   // Sininen KESKIvalo

	Lights.push_back(Light(XYZ(-2.0, -0.0, -4.0), RGB(1.00,0.60,0.15), 0.1)); 
	Lights.push_back(Light(XYZ(+2.0, -2.0, -3.3), RGB(0.25,1.00,0.25), 0.1)); 
	Lights.push_back(Light(XYZ(+2.0, +1.6, -3.3), RGB(0.45,0.25,1.00), 0.1)); 
    
    int Parilliset = 1;
    int Parittomat = 1;
    
    for(int a=1; a<argc; ++a)
    {
        const char *s = argv[a];
        if(!strcmp(s, "2"))Parittomat = 0;
        if(!strcmp(s, "1"))Parilliset = 0;
    }

    if(Parilliset && !Parittomat) ALKUY = YDIM;
    else if(Parittomat && !Parilliset) ALKUY = 0;
    else ALKUY = 0;//(1024-YDIM)/4;

    for(FLOAT ax=1.4; ax<M_PI*4; ax+=M_PI / 2000)
    //FLOAT ax;
    //for(ax=0, w=-1.3; w<=1.3; w+=0.01)
    {
        //Matrix eyematrix(ax-M_PI, -SIN(ax/2), -0.3 + SIN(ax));
        //Matrix eyematrix(0.1, -0.7, 0);
        Matrix eyematrix(M_PI/2+ax, M_PI/4+ax, 0);
        //Matrix eyematrix(0, 0, 0);
        //Matrix lightmatrix(ax*2, COS(ax)-0.9, COS(ax));
        Matrix lightmatrix = eyematrix;
        
        static unsigned imgind=1000;
        static const unsigned char tgaBuf[] = {0,0,2, 0,0, 0,0, 0,  0,0, 0,0, XDIM&255, XDIM>>8, YDIM&255, YDIM>>8, 24,32};
        char fnBuf[64];
        sprintf(fnBuf, "tga/mand%04u.tga", ++imgind);
        
        //if(imgind < 1603)continue;
        
        if((imgind & 1)==1 && !Parittomat)continue;
        if((imgind & 1)==0 && !Parilliset)continue;
        
        FILE *fp = fopen(fnBuf, "wb");
        fwrite(tgaBuf, 1, sizeof tgaBuf, fp);

        XYZ beam(0, 0, -EYEDIST);
        beam.Rotate(eyematrix);
	    beam.x -= 0.3;
	    
	    time_t begin = time(NULL);
        
        for(unsigned l=0; l<Lights.size(); ++l)
        {
            Lights[l].pos = Lights[l].oldpos;
            Lights[l].pos.Rotate(lightmatrix);
        }

        for(unsigned y=0; y<YDIM; ++y)
        {
        	printf("\r%.1f%%", y*100.0/YDIM); fflush(stdout);
        	
            const FLOAT ydisp = (y / (FLOAT)(YDIM-1) - 0.5) * (PROJSIZEY*2.0);
            
            for(unsigned x=0; x<XDIM; ++x)
            {
                const FLOAT xdisp = (x / (FLOAT)(XDIM-1) - 0.5) * (PROJSIZEX*2.0);
                
                XYZ dir(xdisp,ydisp,PROJDIST);
                dir.SetLength(BEAMSPEED);
                dir.Rotate(eyematrix);
                
#if DISPLAYRENDER
                PSet(x,y, 15);
#endif
                RGB pix = RayTrace(beam, dir, MIRRORCOUNT);

                const unsigned char R = pix.getR();
                const unsigned char G = pix.getG();
                const unsigned char B = pix.getB();
                
                putc(B, fp);
                putc(G, fp);
                putc(R, fp);
#if DISPLAYRENDER
                unsigned char c = (B*5/256)
                                + (R*5/256) * 5
                                + (G*9/256) * (5*5);
                PSet(x,y, 16 + c);
#endif
            }
        }
        
        fclose(fp);
        
        time_t end = time(NULL);
        
        printf("Next, took %u seconds\n", end-begin); fflush(stdout);
    }
    return 0;
}
