#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#define XDIM 640
#define YDIM 480

static unsigned char Pal[768];
#include <linux/fb.h>
static class Video
{
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        for(int a=0; a<120; a++)
        {
            double mag = a/120.0;
            
            int r = int(mag*511);
            int g = int(mag*511)-256;
            int b = int(mag*2400)-(2400-256);
            
            if(r>255)r=255;if(r<0)r=0;
            if(g>255)g=255;if(g<0)g=0;
            if(b>255)b=255;if(b<0)b=0;
            
            cmap.red  [a+16] = r * 65535 / 255;
            cmap.green[a+16] = g * 65535 / 255;
            cmap.blue [a+16] = b * 65535 / 255;
            
            cmap.red  [a+16+120] = b * 65535 / 255;
            cmap.green[a+16+120] = g * 65535 / 255;
            cmap.blue [a+16+120] = r * 65535 / 255;
        }
        for(int a=0; a<256; ++a)
        {
        	Pal[a*3+0] = cmap.blue [a]/256;
        	Pal[a*3+1] = cmap.green[a]/256;
        	Pal[a*3+2] = cmap.red  [a]/256;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        /*p[y*xres+x] = c;*/
        x+=(1280-XDIM*2)/4; y+=(1024-YDIM*2)/4;
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
    }
    unsigned char read(unsigned x, unsigned y)
    {
        /*return p[y*xres+x];*/
        x+=(1280-XDIM*2)/4; y+=(1024-YDIM*2)/4;
        return p[(y*2  )*xres+(x*2  )];
    }
} PSet;

#define MaxIter 110

#define MANDEL 1

char CalcPix(double RP,double IP, double RPart,double IPart)
{
    register double r = RP, i = IP;
    int c;
    for(c=0; ; )
    {
        const double r2 = r*r;
        const double i2 = i*i;
        
        if(r2+i2 >= 4)
        {
        	break;
            /* Would never return back to <4. Not part of set. Exit. */
        }
        
        i = r*i*2 + (MANDEL ? IP : IPart);
        r = r2-i2 + (MANDEL ? RP : RPart);
        
        if(++c >= MaxIter)
        {
        	return 120 + int((r2+i2) * 120 / 4);
            /* Tried enough. We are inside. See how inside. */
        }
    }
    return c;
}

#include <termios.h>
static int getch()
{
	struct termio term, back;
	int c;
	ioctl(0, TCGETA, &term);
	ioctl(0, TCGETA, &back);
	term.c_lflag &= ~(ECHO | ICANON);
	term.c_cc[VMIN] = 1;
	ioctl(0, TCSETA, &term);
	read(0, &c, 1);
	ioctl(0, TCSETA, &back);
	return c;
}

int main()
{
    double XMin,XMax,YMin,YMax, RP,IP;
    
    unsigned char Buf[18] = {0,0,2,0,0,0,0, 0,0,0,0,0, 640&255, 640>>8, 480&255, 480>>8, '', ' '};
    write(2, Buf, sizeof Buf);

    unsigned sx,sy;

    double RPart = -0.282;
    double IPart = +0.682;

    XMin = -2.5; XMax = +1.5;
    YMin = -2.5; YMax = +2.5;
    
    XMin=-0.7002; YMin=-0.80396;
    XMax=-0.4998; YMax=-0.60356;
	
    char a;
    do{
        IP = YMin;
        for(sy=0; sy<YDIM; sy++)
        {
            RP = XMin;
            for(sx=0; sx<XDIM; sx++)
            {
            	int c = CalcPix(RP,IP, RPart,IPart);
            	
            	c = 16 + c%120;
            	
            	{
	            	unsigned char R = Pal[(c)*3+0];
	            	unsigned char G = Pal[(c)*3+1];
	            	unsigned char B = Pal[(c)*3+2];
    	        	write(2, &R, 1);
        	    	write(2, &G, 1);
            		write(2, &B, 1);
	            }
	            
	            PSet(sx,sy, c);
	            
                RP += (XMax-XMin)/XDIM;
            }
            IP += (YMax-YMin)/YDIM;
        }
		unsigned WXMin, WYMin, WXMax, WYMax;

		WXMin=WYMin=0;
		WYMax=127;
		WXMax=127 * (YMax-YMin) / (XMax-XMin);

        static char Save[XDIM][4];

        a = 0;
        do{
            for(sx=WXMin; sx<=WXMax; sx++)
            {
                Save[sx][0] = PSet.read(sx,WYMin);
                Save[sx][1] = PSet.read(sx,WYMax);
                PSet(sx,WYMin, rand()%256);
                PSet(sx,WYMax, rand()%256);
            }
            for(sy=WYMin+1; sy<WYMax-1; sy++)
            {
                Save[sy][2] = PSet.read(WXMin,sy);
                Save[sy][3] = PSet.read(WXMax,sy);
                PSet(WXMin,sy, rand()%256);
                PSet(WXMax,sy, rand()%256);
            }

            char b = getch();

            for(sx=WXMin; sx<=WXMax; sx++)
            {
                PSet(sx,WYMin, Save[sx][0]);
                PSet(sx,WYMax, Save[sx][1]);
            }
            for(sy=WYMin+1; sy<WYMax-1; sy++)
            {
                PSet(WXMin,sy, Save[sy][2]);
                PSet(WXMax,sy, Save[sy][3]);
            }

            int Vauhti = 1+(MaxIter-PSet.read((WXMin+WXMax)/2, (WYMin+WYMax)/2))/10;
            switch(b)
            {
                case '': if(WYMax>=Vauhti){WYMin-=Vauhti; WYMax-=Vauhti;}; break;
                case '': if(WYMax+Vauhti<YDIM){WYMin+=Vauhti; WYMax+=Vauhti;}; break;
                case '': if(WXMax>=Vauhti){WXMin-=Vauhti; WXMax-=Vauhti;}; break;
                case '': if(WXMax+Vauhti<XDIM){WXMin+=Vauhti; WXMax+=Vauhti;}; break;
                case '-': a = 2;
                case 27: a = 2; break;
                case 10:
                case 13: a = 1;
            }
        }while(!a);
        double TX1 = (WXMin)*(XMax-XMin)/(float)XDIM+XMin;
        double TY1 = (WYMin)*(YMax-YMin)/(float)YDIM+YMin;
        double TX2 = (WXMax)*(XMax-XMin)/(float)XDIM+XMin;
        double TY2 = (WYMax)*(YMax-YMin)/(float)YDIM+YMin;
        XMin = TX1; YMin = TY1;
        XMax = TX2; YMax = TY2;
    }while(a != 2);
    return 0;
}
